@@ -2,35 +2,36 @@ import os
 import sys
 import logging
 import base64
-from xml.etree import ElementTree
 from pathlib import Path
+import platform
+from xml.dom.minidom import parse
 
 from qt_material.resources import ResourseGenerator, RESOURCES_PATH
 GUI = True
 
 if 'PySide2' in sys.modules:
     from PySide2.QtGui import QFontDatabase, QColor, QGuiApplication, QPalette
-    from PySide2.QtWidgets import QAction, QColorDialog
+    from PySide2.QtWidgets import QAction, QColorDialog, QActionGroup
     from PySide2.QtUiTools import QUiLoader
     from PySide2.QtCore import Qt, QDir
 
 elif 'PySide6' in sys.modules:
-    from PySide6.QtGui import QFontDatabase, QAction, QColor, QGuiApplication, QPalette
+    from PySide6.QtGui import QFontDatabase, QAction, QColor, QGuiApplication, QPalette, QActionGroup
     from PySide6.QtWidgets import QColorDialog
     from PySide6.QtUiTools import QUiLoader
     from PySide6.QtCore import Qt, QDir
 
 elif 'PyQt5' in sys.modules:
     from PyQt5.QtGui import QFontDatabase, QColor, QGuiApplication, QPalette
-    from PyQt5.QtWidgets import QAction, QColorDialog
+    from PyQt5.QtWidgets import QAction, QColorDialog, QActionGroup
     from PyQt5.QtCore import Qt, QDir
     from PyQt5 import uic
 
 elif 'PyQt6' in sys.modules:
-    from PyQt5.QtGui import QFontDatabase, QColor, QGuiApplication, QPalette
-    from PyQt5.QtWidgets import QAction, QColorDialog
-    from PyQt5.QtCore import Qt, QDir
-    from PyQt5 import uic
+    from PyQt6.QtGui import QFontDatabase, QColor, QGuiApplication, QPalette, QAction, QActionGroup
+    from PyQt6.QtWidgets import QColorDialog
+    from PyQt6.QtCore import Qt, QDir
+    from PyQt6 import uic
 else:
     GUI = False
     logging.warning("qt_material must be imported after PySide or PyQt!")
@@ -46,8 +47,10 @@ def export_theme(theme='', qss=None, rcc=None, invert_secondary=False, extra={},
     if not os.path.isabs(output) and not output.startswith('.'):
             output = f'.{output}'
 
-    stylesheet = build_stylesheet(
-        theme, invert_secondary, extra, output)
+    stylesheet = build_stylesheet(theme, invert_secondary, extra, output)
+
+    if output.startswith('.'):
+        output = output[1:]
 
     with open(qss, 'w') as file:
         file.writelines(stylesheet.replace('icon:/', prefix))
@@ -58,9 +61,6 @@ def export_theme(theme='', qss=None, rcc=None, invert_secondary=False, extra={},
             file.write('<RCC>\n')
             file.write(f'  <qresource prefix="{prefix[:-2]}">\n')
 
-            if output.startswith('.'):
-                output = output[1:]
-
             for subfolder in ['disabled', 'primary']:
                 files = os.listdir(os.path.join(
                     os.path.abspath(output), subfolder))
@@ -69,12 +69,12 @@ def export_theme(theme='', qss=None, rcc=None, invert_secondary=False, extra={},
                     file.write(
                         f'    <file>{output}/{subfolder}/{filename}</file>\n')
 
-            file.write('  </qresource\n')
+            file.write('  </qresource>\n')
 
             file.write(f'  <qresource prefix="file">\n')
             if qss:
                 file.write(f'    <file>{qss}</file>\n')
-            file.write('  </qresource\n')
+            file.write('  </qresource>\n')
 
             file.write('</RCC>\n')
 
@@ -82,7 +82,10 @@ def export_theme(theme='', qss=None, rcc=None, invert_secondary=False, extra={},
 # ----------------------------------------------------------------------
 def build_stylesheet(theme='', invert_secondary=False, extra={}, parent='theme'):
     """"""
-    add_fonts()
+    try:
+        add_fonts()
+    except Exception as e:
+        logging.warning(e)
 
     theme = get_theme(theme, invert_secondary)
     if theme is None:
@@ -97,6 +100,7 @@ def build_stylesheet(theme='', invert_secondary=False, extra={}, parent='theme')
     theme['icon'] = None
 
     env.filters['opacity'] = opacity
+    env.filters['density'] = density
     # env.filters['as_base64'] = as_base64
     # env.filters['load'] = load
 
@@ -106,16 +110,43 @@ def build_stylesheet(theme='', invert_secondary=False, extra={}, parent='theme')
     theme.setdefault('danger', '#dc3545')
     theme.setdefault('warning', '#ffc107')
     theme.setdefault('success', '#17a2b8')
+    theme.setdefault('density_scale', '0')
+    theme.setdefault('button_shape', 'default')
+    # theme.setdefault('font_size', '13px')
+    # theme.setdefault('line_height', '13px')
 
     theme.update(extra)
 
     if GUI:
         default_palette = QGuiApplication.palette()
-        default_palette.setColor(QPalette.PlaceholderText, QColor(
-            *[int(theme['primaryColor'][i:i + 2], 16) for i in range(1, 6, 2)] + [92]))
-        QGuiApplication.setPalette(default_palette)
+        color = QColor(*[int(theme['primaryColor'][i:i + 2], 16)
+                         for i in range(1, 6, 2)] + [92])
 
-    return stylesheet.render(**theme)
+        try:
+            if hasattr(QPalette, 'PlaceholderText'):  # pyqt5, pyside2, pyside6
+                default_palette.setColor(QPalette.PlaceholderText, color)
+            else:  # pyqt6
+                default_palette.setColor(QPalette.ColorRole.Text, color)
+            QGuiApplication.setPalette(default_palette)
+
+        except:  # pyside6  & snake_case, true_property
+            default_palette.set_color(QPalette.ColorRole.Text, color)
+            QGuiApplication.set_palette(default_palette)
+
+    environ = {
+
+        'linux': platform.system() == 'Linux',
+        'windows': platform.system() == 'Windows',
+        'darwin': platform.system() == 'Darwin',
+
+        'pyqt5': 'PyQt5' in sys.modules,
+        'pyqt6': 'PyQt6' in sys.modules,
+        'pyside2': 'PySide2' in sys.modules,
+        'pyside6': 'PySide6' in sys.modules,
+    }
+
+    environ.update(theme)
+    return stylesheet.render(environ)
 
 
 # ----------------------------------------------------------------------
@@ -137,8 +168,9 @@ def get_theme(theme_name, invert_secondary=False):
         logging.warning(f"{theme} not exist!")
         return None
 
-    tree = ElementTree.parse(theme)
-    theme = {child.attrib['name']: child.text for child in tree.getroot()}
+    document = parse(theme)
+    theme = {child.getAttribute(
+        'name'): child.firstChild.nodeValue for child in document.getElementsByTagName('color')}
 
     for k in theme:
         os.environ[str(k)] = theme[k]
@@ -167,8 +199,12 @@ def add_fonts():
 
     for font_dir in ['roboto']:
         for font in filter(lambda s: s.endswith('.ttf'), os.listdir(os.path.join(fonts_path, font_dir))):
-            QFontDatabase.addApplicationFont(
-                os.path.join(fonts_path, font_dir, font))
+            try:
+                QFontDatabase.addApplicationFont(
+                    os.path.join(fonts_path, font_dir, font))
+            except:  # snake_case, true_property
+                QFontDatabase.add_application_font(
+                    os.path.join(fonts_path, font_dir, font))
 
 
 # ----------------------------------------------------------------------
@@ -176,10 +212,14 @@ def apply_stylesheet(app, theme='', style=None, save_as=None, invert_secondary=F
     """"""
     if style:
         try:
-            app.setStyle(style)
+            try:
+                app.setStyle(style)
+            except:  # snake_case, true_property
+                app.style = style
         except:
             logging.error(f"The style '{style}' does not exist.")
             pass
+
     stylesheet = build_stylesheet(
         theme, invert_secondary, extra, parent)
     if stylesheet is None:
@@ -189,7 +229,10 @@ def apply_stylesheet(app, theme='', style=None, save_as=None, invert_secondary=F
         with open(save_as, 'w') as file:
             file.writelines(stylesheet)
 
-    return app.setStyleSheet(stylesheet)
+    try:
+        app.setStyleSheet(stylesheet)
+    except:
+        app.style_sheet = stylesheet
 
 
 # ----------------------------------------------------------------------
@@ -201,6 +244,27 @@ def opacity(theme, value=0.5):
     return f'rgba({r}, {g}, {b}, {value})'
 
 
+# ----------------------------------------------------------------------
+def density(value, density_scale, border=0, scale=1, density_interval=4):
+    """"""
+    # https://material.io/develop/web/supporting/density
+    if isinstance(value, str) and value.startswith('@'):
+        return value[1:] * scale
+
+    if value == 'unset':
+        return 'unset'
+
+    if isinstance(value, str):
+        value = float(value.replace('px', ''))
+
+    density = (value + (density_interval * int(density_scale)) -
+               (border * 2)) * scale
+
+    if density < 0:
+        density = 0
+    return density
+
+
 # ----------------------------------------------------------------------
 def set_icons_theme(theme, parent='theme'):
     """"""
@@ -210,9 +274,14 @@ def set_icons_theme(theme, parent='theme'):
     resources.generate()
 
     if GUI:
-        QDir.addSearchPath('icon', resources.index)
-        QDir.addSearchPath('qt_material', os.path.join(
-            os.path.dirname(__file__), 'resources'))
+        try:
+            QDir.addSearchPath('icon', resources.index)
+            QDir.addSearchPath('qt_material', os.path.join(
+                os.path.dirname(__file__), 'resources'))
+        except:  # snake_case, true_property
+            QDir.add_search_path('icon', resources.index)
+            QDir.add_search_path('qt_material', os.path.join(
+                os.path.dirname(__file__), 'resources'))
 
 
 # ----------------------------------------------------------------------
@@ -224,40 +293,113 @@ def list_themes():
     return sorted(list(themes))
 
 
+# ----------------------------------------------------------------------
+def deprecated(replace):
+    """"""
+    # ----------------------------------------------------------------------
+    def wrap1(fn):
+        # ----------------------------------------------------------------------
+        def wrap2(*args, **kwargs):
+            logging.warning(
+                f'This function is deprecated, please use "{replace}" instead.')
+            fn(*args, **kwargs)
+        return wrap2
+
+    return wrap1
+
+
 ########################################################################
 class QtStyleTools:
     """"""
-    extra_colors = {}
+    extra_values = {}
 
     # ----------------------------------------------------------------------
+    @deprecated('set_extra')
     def set_extra_colors(self, extra):
         """"""
-        self.extra_colors = extra
+        self.extra_values = extra
+
+    # ----------------------------------------------------------------------
+    def set_extra(self, extra):
+        """"""
+        self.extra_values = extra
 
     # ----------------------------------------------------------------------
     def add_menu_theme(self, parent, menu):
         """"""
-        for theme in ['default'] + list_themes():
+        self.menu_theme_ = menu
+        action_group = QActionGroup(menu)
+        try:
+            action_group.setExclusive(True)
+        except:
+            action_group.exclusive = True
+
+        for i, theme in enumerate(['default'] + list_themes()):
             action = QAction(parent)
-            action.setText(theme)
-            action.triggered.connect(self._wrapper(
-                parent, theme, self.extra_colors, self.update_buttons))
-            menu.addAction(action)
+            # action.triggered.connect(self._wrapper(parent, theme, self.extra_values, self.update_buttons))
+            action.triggered.connect(lambda: self.update_theme_event(parent))
+            try:
+                action.setText(theme)
+                action.setCheckable(True)
+                action.setChecked(not bool(i))
+                action.setActionGroup(action_group)
+                menu.addAction(action)
+                action_group.addAction(action)
+            except:  # snake_case, true_property
+                action.text = theme
+                action.checkable = True
+                action.checked = not bool(i)
+                action.action_group = action_group
+                menu.add_action(action)
+                action_group.add_action(action)
 
     # ----------------------------------------------------------------------
-    def _wrapper(self, parent, theme, extra, callable_):
+    def add_menu_density(self, parent, menu):
         """"""
-        def iner():
-            self._apply_theme(parent, theme, extra, callable_)
-        return iner
+        self.menu_density_ = menu
+        action_group = QActionGroup(menu)
+
+        try:
+            action_group.setExclusive(True)
+        except:
+            action_group.exclusive = True
+
+        for density in map(str, range(-3, 4)):
+            action = QAction(parent)
+            # action.triggered.connect(self._wrapper(parent, density, self.extra_values, self.update_buttons))
+            action.triggered.connect(lambda: self.update_theme_event(parent))
+            try:
+                action.setText(density)
+                action.setCheckable(True)
+                action.setChecked(density == '0')
+                action.setActionGroup(action_group)
+                menu.addAction(action)
+                action_group.addAction(action)
+            except:  # snake_case, true_property
+                action.text = density
+                action.checkable = True
+                action.checked = density == '0'
+                action.action_group = action_group
+                menu.add_action(action)
+                action_group.add_action(action)
+
+        # menu.add_action(action_group)
+
+    # # ----------------------------------------------------------------------
+    # def _wrapper(self, parent, theme, extra, callable_):
+        # """"""
+        # def iner():
+            # self._apply_theme(parent, theme, extra, callable_)
+        # return iner
+
+    # # ----------------------------------------------------------------------
+    # def _apply_theme(self, parent, theme, extra={}, callable_=None):
+        # """"""
+        # self.apply_stylesheet(parent, theme=theme, invert_secondary=theme.startswith(
+            # 'light'), extra=extra, callable_=callable_)
 
     # ----------------------------------------------------------------------
-    def _apply_theme(self, parent, theme, extra={}, callable_=None):
-        """"""
-        self.apply_stylesheet(parent, theme=theme, invert_secondary=theme.startswith(
-            'light'), extra=extra, callable_=callable_)
 
-    # ----------------------------------------------------------------------
     def apply_stylesheet(self, parent, theme, invert_secondary=False, extra={}, callable_=None):
         """"""
         if theme == 'default':
@@ -269,6 +411,25 @@ class QtStyleTools:
         if callable_:
             callable_()
 
+    # ----------------------------------------------------------------------
+    def update_theme_event(self, parent):
+        """"""
+        try:
+            density = [action.text() for action in self.menu_density_.actions(
+            ) if action.isChecked()][0]
+            theme = [action.text() for action in self.menu_theme_.actions()
+                     if action.isChecked()][0]
+        except:
+            density = [
+                action.text for action in self.menu_density_.actions() if action.checked][0]
+            theme = [action.text for action in self.menu_theme_.actions()
+                     if action.checked][0]
+
+        self.extra_values['density_scale'] = density
+
+        self.apply_stylesheet(parent, theme=theme, invert_secondary=theme.startswith(
+            'light'), extra=self.extra_values, callable_=self.update_buttons)
+
     # ----------------------------------------------------------------------
     def update_buttons(self):
         """"""
@@ -283,19 +444,30 @@ class QtStyleTools:
         elif 'dark' in os.environ['QTMATERIAL_THEME']:
             self.dock_theme.checkBox_ligh_theme.setChecked(False)
 
-        if self.dock_theme.checkBox_ligh_theme.isChecked():
-            theme['secondaryColor'], theme['secondaryLightColor'], theme['secondaryDarkColor'] = theme[
-                'secondaryColor'], theme['secondaryDarkColor'], theme['secondaryLightColor']
+        try:
+            if self.dock_theme.checkBox_ligh_theme.isChecked():
+                theme['secondaryColor'], theme['secondaryLightColor'], theme['secondaryDarkColor'] = theme[
+                    'secondaryColor'], theme['secondaryDarkColor'], theme['secondaryLightColor']
+        except:  # snake_case, true_property
+            if self.dock_theme.checkBox_ligh_theme.checked:
+                theme['secondaryColor'], theme['secondaryLightColor'], theme['secondaryDarkColor'] = theme[
+                    'secondaryColor'], theme['secondaryDarkColor'], theme['secondaryLightColor']
 
         for color_ in self.colors:
             button = getattr(self.dock_theme, f'pushButton_{color_}')
 
             color = theme[color_]
 
-            if self.get_color(color).getHsv()[2] < 128:
-                text_color = '#ffffff'
-            else:
-                text_color = '#000000'
+            try:
+                if self.get_color(color).getHsv()[2] < 128:
+                    text_color = '#ffffff'
+                else:
+                    text_color = '#000000'
+            except:  # snake_case, true_property
+                if self.get_color(color).get_hsv()[2] < 128:
+                    text_color = '#ffffff'
+                else:
+                    text_color = '#000000'
 
             button.setStyleSheet(f"""
             *{{
@@ -326,9 +498,13 @@ class QtStyleTools:
                 <color name="secondaryTextColor">{secondaryTextColor}</color>
               </resources>
             """.format(**self.custom_colors))
-        light = self.dock_theme.checkBox_ligh_theme.isChecked()
+        try:
+            light = self.dock_theme.checkBox_ligh_theme.isChecked()
+        except:  # snake_case, true_property
+            light = self.dock_theme.checkBox_ligh_theme.checked
+
         self.apply_stylesheet(parent, 'my_theme.xml', invert_secondary=light,
-                              extra=self.extra_colors, callable_=self.update_buttons)
+                              extra=self.extra_values, callable_=self.update_buttons)
 
     # ----------------------------------------------------------------------
     def set_color(self, parent, button_):
@@ -336,15 +512,31 @@ class QtStyleTools:
         def iner():
             initial = self.get_color(self.custom_colors[button_])
             color_dialog = QColorDialog(parent=parent)
-            color_dialog.setCurrentColor(initial)
+            try:
+                color_dialog.setCurrentColor(initial)
+            except:  # snake_case, true_property
+                color_dialog.current_color = initial
             done = color_dialog.exec_()
-            color_ = color_dialog.currentColor()
 
-            if done and color_.isValid():
-                color = '#' + ''.join([hex(v)[2:].ljust(2, '0')
-                                       for v in color_.toTuple()[:3]])
-                self.custom_colors[button_] = color
-                self.update_theme(parent)
+            try:
+                color_ = color_dialog.currentColor()
+            except:  # snake_case, true_property
+                color_ = color_dialog.current_color
+
+            try:
+                if done and color_.isValid():
+                    rgb_255 = [color_.red(), color_.green(), color_.blue()]
+                    color = '#' + ''.join([hex(v)[2:].ljust(2, '0')
+                                           for v in rgb_255])
+                    self.custom_colors[button_] = color
+                    self.update_theme(parent)
+            except:  # snake_case, true_property
+                if done and color_.is_valid():
+                    rgb_255 = [color_.red(), color_.green(), color_.blue()]
+                    color = '#' + ''.join([hex(v)[2:].ljust(2, '0')
+                                           for v in rgb_255])
+                    self.custom_colors[button_] = color
+                    self.update_theme(parent)
 
         return iner
 
@@ -365,12 +557,18 @@ class QtStyleTools:
         if 'PySide2' in sys.modules or 'PySide6' in sys.modules:
             self.dock_theme = QUiLoader().load(os.path.join(
                 os.path.dirname(__file__), 'dock_theme.ui'))
-        elif 'PyQt5' in sys.modules:
+        elif 'PyQt5' in sys.modules or 'PyQt6' in sys.modules:
             self.dock_theme = uic.loadUi(os.path.join(
                 os.path.dirname(__file__), 'dock_theme.ui'))
 
-        parent.addDockWidget(Qt.LeftDockWidgetArea, self.dock_theme)
-        self.dock_theme.setFloating(True)
+        try:
+            parent.addDockWidget(
+                Qt.DockWidgetArea.LeftDockWidgetArea, self.dock_theme)
+            self.dock_theme.setFloating(True)
+        except:  # snake_case, true_property
+            parent.add_dock_widget(
+                Qt.DockWidgetArea.LeftDockWidgetArea, self.dock_theme)
+            self.dock_theme.floating = True
 
         self.update_buttons()
         self.dock_theme.checkBox_ligh_theme.clicked.connect(
